@@ -35,24 +35,23 @@ CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 #define FEH_JITTER_OFFSET 2
 #define FEH_JITTER_TIME 1
 
-fehbb buttons;
+extern fehkb keys;
 
 feh_event_handler *ev_handler[LASTEvent];
 
 static void feh_event_handle_ButtonPress(XEvent * ev);
 static void feh_event_handle_ButtonRelease(XEvent * ev);
-static void feh_event_handle_ConfigureNotify(XEvent * ev);
 static void feh_event_handle_LeaveNotify(XEvent * ev);
 static void feh_event_handle_MotionNotify(XEvent * ev);
 static void feh_event_handle_ClientMessage(XEvent * ev);
 
-static void feh_set_bb(fehbutton *bb, int modifier, char button)
+static void feh_set_bb(fehkey *bb, int modifier, char button)
 {
-	bb->modifier = modifier;
+	bb->state  = modifier;
 	bb->button = button;
 }
 
-static void feh_set_parse_bb_partial(fehbutton *button, char *binding)
+static void feh_set_parse_bb_partial(fehkey *button, char *binding)
 {
 	char *cur = binding;
 	int mod = 0;
@@ -84,9 +83,13 @@ static void feh_set_parse_bb_partial(fehbutton *button, char *binding)
 	}
 
 	button->button = atoi(cur);
-	button->modifier = mod;
+	button->state  = mod;
 }
 
+/*
+ * Called after init_keyevents in keyevents.c
+ * -> no need to memset
+ */
 void init_buttonbindings(void)
 {
 	char *home = NULL;
@@ -94,20 +97,17 @@ void init_buttonbindings(void)
 	char *confpath = NULL;
 	char line[128];
 	char action[32], button[8];
-	struct __fehbutton *cur_bb = NULL;
+	struct __fehkey *cur_bb = NULL;
 	FILE *conf = NULL;
 	int read = 0;
 
-	memset(&buttons, 0, sizeof(buttons));
-
-	feh_set_bb(&buttons.reload, 0, 0);
-	feh_set_bb(&buttons.pan,    0, 1);
-	feh_set_bb(&buttons.zoom,   0, 2);
-	feh_set_bb(&buttons.menu,   0, 3);
-	feh_set_bb(&buttons.prev,   0, 4);
-	feh_set_bb(&buttons.next,   0, 5);
-	feh_set_bb(&buttons.blur,   4, 1);
-	feh_set_bb(&buttons.rotate, 4, 2);
+	feh_set_bb(&keys.pan,         0, 1);
+	feh_set_bb(&keys.zoom,        0, 2);
+	feh_set_bb(&keys.toggle_menu, 0, 3);
+	feh_set_bb(&keys.prev_img,    0, 4);
+	feh_set_bb(&keys.next_img,    0, 5);
+	feh_set_bb(&keys.blur,        4, 1);
+	feh_set_bb(&keys.rotate,      4, 2);
 
 	home = getenv("HOME");
 	confhome = getenv("XDG_CONFIG_HOME");
@@ -129,44 +129,52 @@ void init_buttonbindings(void)
 	while (fgets(line, sizeof(line), conf)) {
 		*action = '\0';
 		*button = '\0';
+		cur_bb = NULL;
 
 		read = sscanf(line, "%31s %7s\n", (char *) &action, (char *) &button);
 
 		if ((read == EOF) || (read == 0) || (line[0] == '#'))
 			continue;
 
+		/*
+		 * Note: This isn't really good code. But it works, and since it only
+		 * runs once for each button config line the runtime penalty compared to
+		 * e.g. a hash table is negligible in this case.
+		 */
 		if (!strcmp(action, "reload"))
-			cur_bb = &buttons.reload;
+			cur_bb = &keys.reload;
 		else if (!strcmp(action, "pan"))
-			cur_bb = &buttons.pan;
+			cur_bb = &keys.pan;
 		else if (!strcmp(action, "zoom"))
-			cur_bb = &buttons.zoom;
-		else if (!strcmp(action, "menu"))
-			cur_bb = &buttons.menu;
-		else if (!strcmp(action, "prev"))
-			cur_bb = &buttons.prev;
-		else if (!strcmp(action, "next"))
-			cur_bb = &buttons.next;
+			cur_bb = &keys.zoom;
+		else if (!strcmp(action, "menu") || !strcmp(action, "toggle_menu"))
+			cur_bb = &keys.toggle_menu;
+		else if (!strcmp(action, "prev") || !strcmp(action, "prev_img"))
+			cur_bb = &keys.prev_img;
+		else if (!strcmp(action, "next") || !strcmp(action, "next_img"))
+			cur_bb = &keys.next_img;
 		else if (!strcmp(action, "blur"))
-			cur_bb = &buttons.blur;
+			cur_bb = &keys.blur;
 		else if (!strcmp(action, "rotate"))
-			cur_bb = &buttons.rotate;
+			cur_bb = &keys.rotate;
 		else if (!strcmp(action, "zoom_in"))
-			cur_bb = &buttons.zoom_in;
+			cur_bb = &keys.zoom_in;
 		else if (!strcmp(action, "zoom_out"))
-			cur_bb = &buttons.zoom_out;
+			cur_bb = &keys.zoom_out;
 		else
-			weprintf("buttons: Invalid action: %s", action);
+			cur_bb = feh_str_to_kb(action);
 
 		if (cur_bb)
 			feh_set_parse_bb_partial(cur_bb, button);
+		else
+			weprintf("buttons: Invalid action: %s", action);
 	}
 	fclose(conf);
 }
 
-static short feh_is_bb(fehbutton *bb, int button, int mod)
+static short feh_is_bb(fehkey *bb, unsigned int button, unsigned int mod)
 {
-	if ((bb->modifier == mod) && (bb->button == button))
+	if ((bb->state == mod) && (bb->button == button))
 		return 1;
 	return 0;
 }
@@ -193,7 +201,7 @@ void feh_event_init(void)
 static void feh_event_handle_ButtonPress(XEvent * ev)
 {
 	winwidget winwid = NULL;
-	int state, button;
+	unsigned int state, button;
 
 	/* get the heck out if it's a mouse-click on the
 	   cover, we'll hide the menus on release */
@@ -209,23 +217,23 @@ static void feh_event_handle_ButtonPress(XEvent * ev)
 	state = ev->xbutton.state & (ControlMask | ShiftMask | Mod1Mask | Mod4Mask);
 	button = ev->xbutton.button;
 
-	if (!opt.no_menus && feh_is_bb(&buttons.menu, button, state)) {
+	if (!opt.no_menus && feh_is_bb(&keys.toggle_menu, button, state)) {
 		D(("Menu Button Press event\n"));
 		winwidget_show_menu(winwid);
 
-	} else if (feh_is_bb(&buttons.rotate, button, state)
+	} else if (feh_is_bb(&keys.rotate, button, state)
 		   && (winwid->type != WIN_TYPE_THUMBNAIL)) {
 		opt.mode = MODE_ROTATE;
 		winwid->mode = MODE_ROTATE;
 		D(("rotate starting at %d, %d\n", ev->xbutton.x, ev->xbutton.y));
 
-	} else if (feh_is_bb(&buttons.blur, button, state)
+	} else if (feh_is_bb(&keys.blur, button, state)
 		   && (winwid->type != WIN_TYPE_THUMBNAIL)) {
 		opt.mode = MODE_BLUR;
 		winwid->mode = MODE_BLUR;
 		D(("blur starting at %d, %d\n", ev->xbutton.x, ev->xbutton.y));
 
-	} else if (feh_is_bb(&buttons.pan, button, state)) {
+	} else if (feh_is_bb(&keys.pan, button, state)) {
 		D(("Next button, but could be pan mode\n"));
 		opt.mode = MODE_NEXT;
 		winwid->mode = MODE_NEXT;
@@ -234,7 +242,7 @@ static void feh_event_handle_ButtonPress(XEvent * ev)
 		winwid->click_offset_y = ev->xbutton.y - winwid->im_y;
 		winwid->click_start_time = time(NULL);
 
-	} else if (feh_is_bb(&buttons.zoom, button, state)) {
+	} else if (feh_is_bb(&keys.zoom, button, state)) {
 		D(("Zoom Button Press event\n"));
 		opt.mode = MODE_ZOOM;
 		winwid->mode = MODE_ZOOM;
@@ -249,7 +257,7 @@ static void feh_event_handle_ButtonPress(XEvent * ev)
 		winwid->im_click_offset_y = (winwid->click_offset_y
 				- winwid->im_y) / winwid->old_zoom;
 
-	} else if (feh_is_bb(&buttons.zoom_in, button, state)) {
+	} else if (feh_is_bb(&keys.zoom_in, button, state)) {
 		D(("Zoom_In Button Press event\n"));
 		D(("click offset is %d,%d\n", ev->xbutton.x, ev->xbutton.y));
 		winwid->click_offset_x = ev->xbutton.x;
@@ -277,7 +285,7 @@ static void feh_event_handle_ButtonPress(XEvent * ev)
 		winwidget_sanitise_offsets(winwid);
 		winwidget_render_image(winwid, 0, 0);
 
-	} else if (feh_is_bb(&buttons.zoom_out, button, state)) {
+	} else if (feh_is_bb(&keys.zoom_out, button, state)) {
 		D(("Zoom_Out Button Press event\n"));
 		D(("click offset is %d,%d\n", ev->xbutton.x, ev->xbutton.y));
 		winwid->click_offset_x = ev->xbutton.x;
@@ -305,22 +313,23 @@ static void feh_event_handle_ButtonPress(XEvent * ev)
 		winwidget_sanitise_offsets(winwid);
 		winwidget_render_image(winwid, 0, 0);
 
-	} else if (feh_is_bb(&buttons.reload, button, state)) {
+	} else if (feh_is_bb(&keys.reload, button, state)) {
 		D(("Reload Button Press event\n"));
 			feh_reload_image(winwid, 0, 1);
 
-	} else if (feh_is_bb(&buttons.prev, button, state)) {
+	} else if (feh_is_bb(&keys.prev_img, button, state)) {
 		D(("Prev Button Press event\n"));
 		if (winwid->type == WIN_TYPE_SLIDESHOW)
 			slideshow_change_image(winwid, SLIDE_PREV, 1);
 
-	} else if (feh_is_bb(&buttons.next, button, state)) {
+	} else if (feh_is_bb(&keys.next_img, button, state)) {
 		D(("Next Button Press event\n"));
 		if (winwid->type == WIN_TYPE_SLIDESHOW)
 			slideshow_change_image(winwid, SLIDE_NEXT, 1);
 
 	} else {
 		D(("Received other ButtonPress event\n"));
+		feh_event_handle_generic(winwid, state, NoSymbol, button);
 	}
 	return;
 }
@@ -328,8 +337,8 @@ static void feh_event_handle_ButtonPress(XEvent * ev)
 static void feh_event_handle_ButtonRelease(XEvent * ev)
 {
 	winwidget winwid = NULL;
-	int state = ev->xbutton.state & (ControlMask | ShiftMask | Mod1Mask | Mod4Mask);
-	int button = ev->xbutton.button;
+	unsigned int state = ev->xbutton.state & (ControlMask | ShiftMask | Mod1Mask | Mod4Mask);
+	unsigned int button = ev->xbutton.button;
 
 	if (menu_root) {
 		/* if menus are open, close them, and execute action if needed */
@@ -354,7 +363,7 @@ static void feh_event_handle_ButtonRelease(XEvent * ev)
 		return;
 	}
 
-	if (feh_is_bb(&buttons.pan, button, state)) {
+	if (feh_is_bb(&keys.pan, button, state)) {
 		if (opt.mode == MODE_PAN) {
 			D(("Disabling pan mode\n"));
 			opt.mode = MODE_NORMAL;
@@ -392,13 +401,13 @@ static void feh_event_handle_ButtonRelease(XEvent * ev)
 			winwid->mode = MODE_NORMAL;
 		}
 
-	} else if (feh_is_bb(&buttons.rotate, button, state)
-			|| feh_is_bb(&buttons.zoom, button, state)) {
+	} else if (feh_is_bb(&keys.rotate, button, state)
+			|| feh_is_bb(&keys.zoom, button, state)) {
 		D(("Disabling mode\n"));
 		opt.mode = MODE_NORMAL;
 		winwid->mode = MODE_NORMAL;
 
-		if ((feh_is_bb(&buttons.zoom, button, state))
+		if ((feh_is_bb(&keys.zoom, button, state))
 				&& (ev->xbutton.x == winwid->click_offset_x)
 				&& (ev->xbutton.y == winwid->click_offset_y)) {
 			winwid->zoom = 1.0;
@@ -408,7 +417,7 @@ static void feh_event_handle_ButtonRelease(XEvent * ev)
 
 		winwidget_render_image(winwid, 0, 0);
 
-	} else if (feh_is_bb(&buttons.blur, button, state)) {
+	} else if (feh_is_bb(&keys.blur, button, state)) {
 		D(("Disabling Blur mode\n"));
 		opt.mode = MODE_NORMAL;
 		winwid->mode = MODE_NORMAL;
@@ -416,7 +425,7 @@ static void feh_event_handle_ButtonRelease(XEvent * ev)
 	return;
 }
 
-static void feh_event_handle_ConfigureNotify(XEvent * ev)
+void feh_event_handle_ConfigureNotify(XEvent * ev)
 {
 	while (XCheckTypedWindowEvent(disp, ev->xconfigure.window, ConfigureNotify, ev));
 	if (!menu_root) {
@@ -639,7 +648,7 @@ static void feh_event_handle_MotionNotify(XEvent * ev)
 				winwid->im_h = gib_imlib_image_get_height(temp);
 				gib_imlib_free_image_and_decache(temp);
 				if (!winwid->full_screen && !opt.geom_flags)
-					winwidget_resize(winwid, winwid->im_w, winwid->im_h);
+					winwidget_resize(winwid, winwid->im_w, winwid->im_h, 0);
 				winwid->has_rotated = 1;
 			}
 			winwid->im_angle = (ev->xmotion.x - winwid->w / 2) / ((double) winwid->w / 2) * 3.1415926535;
