@@ -35,6 +35,7 @@ CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 /* TODO s/bit/lot */
 void init_index_mode(void)
 {
+	Imlib_Load_Error err;
 	Imlib_Image im_main;
 	Imlib_Image im_temp;
 	int w = 800, h = 600, ww = 0, hh = 0, www, hhh, xxx, yyy;
@@ -167,7 +168,7 @@ void init_index_mode(void)
 	if (!opt.title)
 		s = estrdup(PACKAGE " [index mode]");
 	else
-		s = estrdup(feh_printf(opt.title, NULL));
+		s = estrdup(feh_printf(opt.title, NULL, NULL));
 
 	if (opt.display) {
 		winwid = winwidget_create_from_image(im_main, s, WIN_TYPE_SINGLE);
@@ -302,7 +303,7 @@ void init_index_mode(void)
 		}
 	}
 	if (opt.verbose)
-		putc('\n', stdout);
+		putc('\n', stderr);
 
 	if (opt.title_font) {
 		int fw, fh, fx, fy;
@@ -326,14 +327,17 @@ void init_index_mode(void)
 		else
 			strncpy(output_buf, opt.output_file, 1024);
 
-		gib_imlib_save_image(im_main, output_buf);
-		if (opt.verbose) {
+		gib_imlib_save_image_with_error_return(im_main, output_buf, &err);
+		if (err) {
+			feh_imlib_print_load_error(output_buf, im_main, err);
+		}
+		else if (opt.verbose) {
 			int tw, th;
 
 			tw = gib_imlib_image_get_width(im_main);
 			th = gib_imlib_image_get_height(im_main);
-			fprintf(stdout, PACKAGE " - File saved as %s\n", output_buf);
-			fprintf(stdout,
+			fprintf(stderr, PACKAGE " - File saved as %s\n", output_buf);
+			fprintf(stderr,
 				"    - Image is %dx%d pixels and contains %d thumbnails\n", tw, th, thumbnailcount);
 		}
 	}
@@ -425,8 +429,11 @@ void get_index_string_dim(feh_file *file, Imlib_Font fn, int *fw, int *fh)
 	gib_list *line, *lines;
 	int max_w = 0, total_h = 0;
 
-	if (!opt.index_info)
+	if (!opt.index_info) {
+		*fw = 0;
+		*fh = 0;
 		return;
+	}
 
 	/* called with file = NULL in the setup phase.
 	 * We need a fake file, otherwise feh_printf will remove format specifiers,
@@ -462,7 +469,7 @@ void get_index_string_dim(feh_file *file, Imlib_Font fn, int *fw, int *fh)
 
 char *create_index_string(feh_file * file)
 {
-	return feh_printf(opt.index_info, file);
+	return feh_printf(opt.index_info, file, NULL);
 }
 
 char *create_index_title_string(int num, int w, int h)
